#define VRAM ((unsigned char*)0x8800498D)

#define MIN(a, b) ((a)<(b)?(a):(b))
#define MAX(a, b) ((a)>(b)?(a):(b))

#define FLOAT_FIXED(val)	((int)((val)*4096.0f))

#define FIXED_INT(val)	    ((val)>>12)
#define INT_FIXED(val)	    ((val)<<12)

typedef int int32_t;
typedef long int64_t;

static inline int64_t imul64(int64_t a, int64_t b) { return (a*b) >> 12; }
static inline int32_t imul(int32_t a, int32_t b) { return imul64(a, b); }

extern const unsigned char matrix[4][4];

static inline void clear_vram() {
  int i;
  volatile unsigned char *vram = VRAM;
  for (i = 0; i < 1024; i++) {
    vram[i] = 0;
  }
}

static inline void display_vram() {
	volatile unsigned char *LCD_register_selector = (char*)0xB4000000, *LCD_data_register = (char*)0xB4010000, *vram;
	int i, j;
  vram = VRAM;
	for (i = 0; i < 64; i++) {
		*LCD_register_selector = 4;
		*LCD_data_register = i|192;
		*LCD_register_selector = 4;
		*LCD_data_register = 0;
		*LCD_register_selector = 7;
		for (j = 0; j < 16; j++) *LCD_data_register = *vram++;
	}
}

static inline void dpixel(int x, int y, unsigned char color) {
    if (color > matrix[x&3][y&3]) *(VRAM + (y<<4)+(x>>3)) |= 128>>(x&7);
}

static inline void pixel(int x, int y) {
    *(VRAM + (y<<4)+(x>>3)) |= 128>>(x&7);
}

static unsigned xorshift_state = 69420;
static inline int mrand() {
    unsigned int x = xorshift_state;
    x ^= x << 13;
    x ^= x >> 17;
    x ^= x << 5;
    return xorshift_state = x;
}

static inline void effect() {
#if 0
  int x, y;
	for(y = 0; y < 64; y++) {
		for(x = 0; x < 128; x++) {
			int i;
			for(i = 0; i < 6; i++) {
				int t,c,o,f;
				t = (ticks >> (1+i));
				o = 0;
				f = MAX(0, (127-t));
				c = (((x+o)*f) ^ ((y+o)*f)) >> 11;

				if(c & 1) {
					dpixel(x, y, 40+((i<<6)));
					break;
				}
			}
		}
	}

#else
  static volatile int fern_x = 0, fern_y = 0;
  int i;

  for(i = 0; i < 40; i++) {
    int nx, ny, r;

    /* draw point */
    pixel((fern_y)/330, 31 - ((fern_x)/280));

    /* next point */
    r = (mrand() & (4095));

    if (r < FLOAT_FIXED(0.01f)) {
      nx = 0;
      ny = imul(FLOAT_FIXED(0.16f), fern_y);
    } else if(r < FLOAT_FIXED(0.86f)) {
      nx = imul(FLOAT_FIXED( 0.85f), fern_x) + imul(FLOAT_FIXED(0.04f), fern_y);
      ny = imul(FLOAT_FIXED(-0.04f), fern_x) + imul(FLOAT_FIXED(0.85f), fern_y) + FLOAT_FIXED(1.6f);
    } else if(r < FLOAT_FIXED(0.93f)) {
      nx = imul(FLOAT_FIXED(0.20f), fern_x) - imul(FLOAT_FIXED(0.26f), fern_y);
      ny = imul(FLOAT_FIXED(0.23f), fern_x) + imul(FLOAT_FIXED(0.22f), fern_y) + FLOAT_FIXED(1.6f);
    } else {
      nx = imul(FLOAT_FIXED(-0.15f), fern_x) + imul(FLOAT_FIXED(0.28f), fern_y);
      ny = imul(FLOAT_FIXED( 0.26f), fern_x) + imul(FLOAT_FIXED(0.24f), fern_y) + FLOAT_FIXED(0.44f);
    }

    fern_x = nx;
    fern_y = ny;
  }
#endif
}

void main() {
  clear_vram();
  while (1) {
    effect();
    display_vram();
  }
}
